import math
from typing import List

import torch
from ray_utils import RayBundle
from pytorch3d.renderer.cameras import CamerasBase


# Sampler which implements stratified (uniform) point sampling along rays
class StratifiedRaysampler(torch.nn.Module):
    def __init__(
        self,
        cfg
    ):
        super().__init__()

        self.n_pts_per_ray = cfg.n_pts_per_ray
        self.min_depth = cfg.min_depth
        self.max_depth = cfg.max_depth

    def forward(
        self,
        ray_bundle,
    ):
        # TODO (1.4): Compute z values for self.n_pts_per_ray points uniformly sampled between [near, far]
        # z_near = self.min_depth
        # z_far = self.max_depth
        
        # rays_o = ray_bundle.origins
        # rays_d = ray_bundle.directions
        # device = rays_o.device
        
        # # import pdb; pdb.set_trace()
        # nsteps_to_near = (z_near - rays_o[0, -1]) / rays_d[:, -1]
        # nsteps_to_far = (z_far - rays_o[0, -1]) / rays_d[:, -1]
        # rays_inter_near = rays_o + nsteps_to_near.unsqueeze(1) * rays_d
        # rays_inter_far = rays_o + nsteps_to_far.unsqueeze(1) * rays_d
        
        # steps = torch.arange(self.n_pts_per_ray).to(device) / self.n_pts_per_ray
        
        # z_vals = None

        # # TODO (1.4): Sample points from z values
        # sample_points = rays_inter_near[..., None] + (rays_inter_far - rays_inter_near)[..., None] * steps[None, None, ...]
        # # sample_points = rays_inter_near[..., None] + rays_d[..., None] * steps[None, None, ...]
        # sample_points = sample_points.permute(0, 2, 1)
        # z_vals = sample_points[:, :, -1:]
        
        # TODO (1.4): Compute z values for self.n_pts_per_ray points uniformly sampled between [near, far]
        z_near = self.min_depth
        z_far = self.max_depth
        
        rays_o = ray_bundle.origins
        rays_d = ray_bundle.directions
        device = rays_d.device
        n_rays = rays_d.shape[0]
        n_pts = self.n_pts_per_ray
        z_vals = torch.linspace(z_near, z_far, n_pts).to(device)#.unsqueeze(0).unsqueeze(0)
        # z_vals = z_vals.repeat(n_rays, n_pts, 1)
        
        # n_steps = (z_vals.unsqueeze(0) - rays_o[:, -1].unsqueeze(1)) / rays_d[:, -1].unsqueeze(1) # (n_rays, n_pts)
        # import pdb; pdb.set_trace()
        sample_points = rays_o.unsqueeze(1) + rays_d.unsqueeze(1) * z_vals.unsqueeze(0).unsqueeze(-1) # (n_rays, n_pts, 3)

        # import pdb; pdb.set_trace()
        # Return
        return ray_bundle._replace(
            sample_points=sample_points,
            # sample_lengths=z_vals * torch.ones_like(sample_points[..., :1]),
            sample_lengths=(z_vals * torch.ones_like(sample_points[..., 0])).unsqueeze(-1),
        )


sampler_dict = {
    'stratified': StratifiedRaysampler
}